#include "IEventHandler.h"
#include "Compensation.h"
#include "ToneBuffer.h"
#include "ImageOps.h"

#include <atomic>
#include <iostream>
#include <iomanip>
#include <mutex>

class ImageDownloadSupervisor : public iMS::IEventHandler
{
private:
	std::atomic<bool> m_downloading{ true };
	std::atomic<bool> m_error{ false };
public:
	void EventAction(void* sender, const int message, const int param)
	{
		switch (message)
		{
		case (iMS::ImageDownloadEvents::DOWNLOAD_FINISHED): /*std::cout << "Image Download Finished!" << std::endl; */m_downloading.store(false); break;
		case (iMS::ImageDownloadEvents::DOWNLOAD_ERROR):/* std::cout << "Download error on message " << param << std::endl;*/ m_error.store(true); break;
		}
	}
	bool Busy() const { return m_downloading.load(); }
	void Reset() { m_downloading.store(true); m_error.store(false); }
	bool Error() const { return m_error.load(); };
};

class ImageVerifySupervisor : public iMS::IEventHandler
{
private:
	std::atomic<bool> m_verifying{ true };
public:
	void EventAction(void* sender, const int message, const int param)
	{
		switch (message)
		{
		case (iMS::ImageDownloadEvents::VERIFY_SUCCESS): std::cout << "Image Verify Successful!" << std::endl; m_verifying.store(false); break;
		case (iMS::ImageDownloadEvents::VERIFY_FAIL): std::cout << "Image Verify FAILED! First error byte: " << param << std::endl; m_verifying.store(false); break;
		}
	}
	bool Busy() const { return m_verifying.load(); }
	void Reset() { m_verifying.store(true); }
};

class CompensationTableDownloadSupervisor : public iMS::IEventHandler
{
private:
	std::atomic<bool> m_downloading{ true };
	std::atomic<bool> m_error{ false };
public:
	void EventAction(void* sender, const int message, const int param)
	{
		switch (message)
		{
		case (iMS::CompensationEvents::DOWNLOAD_FINISHED): std::cout << "Comp Table Download Finished!" << std::endl; m_downloading.store(false); break;
		case (iMS::CompensationEvents::DOWNLOAD_ERROR): std::cout << "Download error on message " << param << std::endl; m_error.store(true);  break;
		}
	}
	bool Busy() const { return m_downloading.load(); };
	bool Error() const { return m_error.load(); };
};

class CompensationTableVerifySupervisor : public iMS::IEventHandler
{
private:
	std::atomic<bool> m_verify_success{ false };
	std::atomic<bool> m_verifying{ true };
public:
	void EventAction(void* sender, const int message, const int param)
	{
		switch (message)
		{
		case (iMS::CompensationEvents::VERIFY_SUCCESS): std::cout << "Comp Table Verify Successful!" << std::endl; m_verify_success.store(true); m_verifying.store(false); break;
		case (iMS::CompensationEvents::VERIFY_FAIL): std::cout << "Comp Table Verify FAILED!" << std::endl; m_verifying.store(false); break;
		}
	}
	bool Busy() const { return m_verifying.load(); };
	bool Success() const { return m_verify_success.load(); };
};

class ToneBufferDownloadSupervisor : public iMS::IEventHandler
{
private:
	std::atomic<bool> m_downloading{ true };
	std::atomic<bool> m_error{ false };
public:
	void EventAction(void* sender, const int message, const int param)
	{
		switch (message)
		{
		case (iMS::ToneBufferEvents::DOWNLOAD_FINISHED): m_downloading.store(false); break;
		case (iMS::ToneBufferEvents::DOWNLOAD_ERROR): m_error.store(true); break;
		}
	}
	bool Busy() const { return m_downloading.load(); };
	bool Error() const { return m_error.load(); };
};

class PlaybackSupervisor : public iMS::IEventHandler
{
private:
	std::atomic<bool> m_playing{ false };
public:
	void EventAction(void* sender, const int message, const int param)
	{
		switch (message)
		{
		case (iMS::ImagePlayerEvents::IMAGE_STARTED): std::cout << "Image Playback Started" << std::endl; m_playing.store(true); break;
		case (iMS::ImagePlayerEvents::POINT_PROGRESS): std::cout << "Progress: " << param << std::endl; break;
		case (iMS::ImagePlayerEvents::IMAGE_FINISHED): std::cout << "Image Playback Finished" << std::endl; m_playing.store(false); break;
		}
	}
	bool Busy() const { return m_playing.load(); };
};

class SequenceSupervisor : public iMS::IEventHandler
{
private:
	std::atomic<bool> play = false;
	std::atomic<bool> started = false;
	std::atomic<bool> tone = false;
	std::atomic<int> position_index = -1;
	std::mutex uuid_mutex;
	std::array<uint8_t, 16> uuid;
public:
	void EventAction(void* sender, const int message, const int param, const std::vector<uint8_t> data)
	{
		switch (message)
		{
		case (iMS::SequenceEvents::SEQUENCE_START): {
			{
				// Thread safe update to the uuid member
				std::lock_guard<std::mutex> lock(uuid_mutex);
				std::copy_n(data.begin(), 16, uuid.begin());
			}
			play.store(true);
			started.store(true);
			break;
		}
		case (iMS::SequenceEvents::SEQUENCE_POSITION): {
			{
				// Thread safe update to the uuid member
				std::lock_guard<std::mutex> lock(uuid_mutex);
				std::copy_n(data.begin(), 16, uuid.begin());
			}
			position_index.store(param);
			break;
		}
		}
	}
	void EventAction(void* sender, const int message, const int param)
	{
		switch (message)
		{
		case (iMS::SequenceEvents::SEQUENCE_FINISHED): std::cout << "Sequence Finished! Last Image Index played was #" << param << std::endl; play.store(false); break;
		case (iMS::SequenceEvents::SEQUENCE_ERROR): std::cout << "Sequence ERROR!" << std::endl; play.store(false); break;
		case (iMS::SequenceEvents::SEQUENCE_TONE): tone.store(true); break;
		}
	}
	bool Playing() { return play.load(); }
	bool SequenceStarted() { 
		bool ret = started.load();
		started.store(false); 
		return ret; 
	}
	std::array<uint8_t, 16> UUID() {
		std::array<uint8_t, 16> uuid_copy;
		{
			// Thread safe update to the local uuid copy
			std::lock_guard<std::mutex> lock(uuid_mutex);
			uuid_copy = uuid;
		}
		return uuid_copy;
	}
	bool ToneStarted() {
		bool ret = tone.load();
		tone.store(false);
		return ret;
	}
	int GetPosition() { 
		int ret = position_index.load(); 
		position_index.store(-1);
		return ret;
	}
};

