// iMSTemplate.cpp : Defines the entry point for the console application.
//

#include "iMS_EnhancedSequences.h"
#include "MySequences.h"
#include "Supervisors.h"

#include <sstream>

/*
 * Uncomment these defines to include the types of interface on which you wish to communicate with the iMS
 */
//#define  ENABLE_SERIAL_PORT_SCAN
#define  ENABLE_ETHERNET_SCAN
#define  ENABLE_USB_SCAN

 /*
 * Uncomment these defines to restrict iMS scan to specific interface ports (using conventional interface type port naming)
 * Note that it is possible to add additional ports by inserting extra lines of "PortMask.push_back(...)" code below
 */
#define SERIAL_PORT_MASK "COM3"
//#define ETHERNET_PORT_MASK  "192.168.2.100"

/*
* Lambda for measuring execution time of a given function
*/
auto measureTime =
[](auto&& func, auto&&... params) {
	// get time before function invocation
	const auto& start = std::chrono::high_resolution_clock::now();
	// function invocation using perfect forwarding
	std::forward<decltype(func)>(func)(std::forward<decltype(params)>(params)...);
	// get time after function invocation
	const auto& stop = std::chrono::high_resolution_clock::now();
	std::cout << " ===> Elapsed Execution Time: " << std::chrono::duration_cast<std::chrono::milliseconds>(stop - start).count() << "ms" << std::endl;
};

// This is a pointer to the system we are going to be working on
static std::unique_ptr < ConnectionList > connList;
static std::shared_ptr < IMSSystem > myiMS;

// Initialise this and use it as single point of access to IMS Control classes
std::unique_ptr < MyIMSAccess > glbl;

std::unique_ptr < MySequences > myseq;

int main(int argc, char* argv[])
{
	char ch;

	std::cout << "   ---  Isomet iMS Enhanced Sequences Example (c) 2020  ---" << std::endl << std::endl;

	//
	// CONNECT TO iMS System
	//
	connList = std::make_unique<ConnectionList>();
	auto& modules = connList->modules();   // returns a ListBase<string>

	// Display all supported modules
	/*for (auto&& mname : modules)
	{
		std::cout << "Module: " << mname << std::endl;
	}*/

	std::cout << "Serial port scan ";
#if !defined(ENABLE_SERIAL_PORT_SCAN)
	std::cout << "disabled" << std::endl;
	if (std::find(modules.begin(), modules.end(), "CM_SERIAL") != modules.end()) connList->config("CM_SERIAL").IncludeInScan = false;
#else
	std::cout << "enabled ";
	// default is enabled.
#if defined(SERIAL_PORT_MASK)
	std::cout << "on port " << SERIAL_PORT_MASK << std::endl;
	if (std::find(modules.begin(), modules.end(), "CM_SERIAL") != modules.end()) connList->config("CM_SERIAL").PortMask.push_back(SERIAL_PORT_MASK);
#else
	std::cout << "on all available ports" << std::endl;
#endif
#endif

	std::cout << "Ethernet port scan ";
#if !defined(ENABLE_ETHERNET_SCAN)
	std::cout << "disabled" << std::endl;
	if (std::find(modules.begin(), modules.end(), "CM_ETH") != modules.end()) connList->config("CM_ETH").IncludeInScan = false;
#else
	std::cout << "enabled ";
	// default is enabled.
#if defined(ETHERNET_PORT_MASK)
	std::cout << "on port " << ETHERNET_PORT_MASK << std::endl;
	if (std::find(modules.begin(), modules.end(), "CM_ETH") != modules.end()) connList->config("CM_ETH").PortMask.push_back(ETHERNET_PORT_MASK);
#else
	std::cout << "on all available ports" << std::endl;
#endif
#endif

	std::cout << "USB port scan ";
#if !defined(ENABLE_USB_SCAN)
	std::cout << "disabled" << std::endl;
	if (std::find(modules.begin(), modules.end(), "CM_USBSS") != modules.end()) connList->config("CM_USBSS").IncludeInScan = false;
	if (std::find(modules.begin(), modules.end(), "CM_USBSS") != modules.end()) connList->config("CM_USBLITE").IncludeInScan = false;
#else
	std::cout << "enabled" << std::endl;
#endif

	// Run full system scan
	std::cout << "Scanning for connected iMS Systems. Please wait. . . ." << std::endl;
	std::vector<IMSSystem> fulliMSList = connList->scan();

	// Initialise the iMS System pointer
	myiMS = std::make_shared<IMSSystem>();

	if (fulliMSList.empty()) {
		std::cout << "Can't connect to an iMS System.  Press a key to exit." << std::endl;
		while (!_kbhit());
		return -1;
	}
	else {
		std::cout << "Discovered " << fulliMSList.size() << " iMS Systems: " << std::endl;
		int ims_count = 1;
		for (std::vector<IMSSystem>::const_iterator it = fulliMSList.cbegin(); it != fulliMSList.cend(); ++it) {
			std::cout << "  " << ims_count++ << ") " << it->ConnPort() << std::endl;
		}
		while (1) {
			std::cout << "Select a system to continue ... ";
			std::cin >> ch;
			// Don't connect more than 9 systems!
			if ((ch - '0') > 0 && (ch - '0') <= (int)(fulliMSList.size())) {
				*myiMS = fulliMSList[ch - '1'];
				break;
			}
		}
		std::cout << std::endl;
	}

	std::cout << "Connecting ... ";
	myiMS->Connect();
	std::cout << "OK" << std::endl;


	// Application Code goes here
	//

	// Create global access object
	glbl = std::make_unique<MyIMSAccess>(myiMS);

	// Create an iMS::ImageGroup filled with character data from free serif font map
	measureTime(createFontMapImageGroup);

	// Download the font map to the iMS hardware
	measureTime(downloadFontMapToIMS, myiMS);

	// View Image Table
	std::cout << "Do you want to view the Image table? [Y/N] " << std::endl;
	std::cin >> ch;
	if (toupper(ch) == 'Y') {
		ImageTableViewer itv(*myiMS);
		std::cout << itv;
	}

	// Download an example Tone Map
	measureTime(createAndDownloadToneMap, myiMS);

	//
	// Set Signal Path conditions
	//

	// For this demo, we will not use LUT compensation
	glbl->sigpath().EnableImagePathCompensation(SignalPath::Compensation::BYPASS, SignalPath::Compensation::BYPASS);

	// Set Maximum amplitude
	glbl->sigpath().UpdateDDSPowerLevel(Percent(100.0));
	glbl->sigpath().SwitchRFAmplitudeControlSource(SignalPath::AmplitudeControl::INDEPENDENT);
	glbl->sigpath().UpdateRFAmplitude(SignalPath::AmplitudeControl::INDEPENDENT, Percent(100.0));

	// Bypass LUT for Local Tone Buffer.
	glbl->sigpath().UpdateLocalToneBuffer(SignalPath::Compensation::BYPASS, SignalPath::Compensation::BYPASS);
	
	// Object that handles all queue management on the iMS
	glbl->seqMgr().QueueClear();

	// Store for all the sequences we create and download
	myseq = std::make_unique<MySequences>(*myiMS);

	SequenceSupervisor seqs;

	glbl->seqMgr().SequenceEventSubscribe(SequenceEvents::SEQUENCE_ERROR, &seqs);
	glbl->seqMgr().SequenceEventSubscribe(SequenceEvents::SEQUENCE_FINISHED, &seqs);
	glbl->seqMgr().SequenceEventSubscribe(SequenceEvents::SEQUENCE_START, &seqs);
	glbl->seqMgr().SequenceEventSubscribe(SequenceEvents::SEQUENCE_TONE, &seqs);
	glbl->seqMgr().SequenceEventSubscribe(SequenceEvents::SEQUENCE_POSITION, &seqs);

	std::cout << std::endl << "   >>> DEMO is Ready.  Press any key to begin.. <<<" << std::endl;
	while (!_kbhit());
	_getch();

	SequenceMenuInit();
	while (1) {
		DrawQueue();
		UpdateStatus(seqs);
		DrawStatus(seqs);
		if (-1 == RunMenu())
			break;
	}

	//
	// CLEANUP
	//
	glbl->seqMgr().Stop(ImagePlayer::StopStyle::IMMEDIATELY);
	glbl->seqMgr().QueueClear();
	glbl->seqMgr().SequenceEventUnsubscribe(SequenceEvents::SEQUENCE_ERROR, &seqs);
	glbl->seqMgr().SequenceEventUnsubscribe(SequenceEvents::SEQUENCE_FINISHED, &seqs);
	glbl->seqMgr().SequenceEventUnsubscribe(SequenceEvents::SEQUENCE_START, &seqs);
	glbl->seqMgr().SequenceEventUnsubscribe(SequenceEvents::SEQUENCE_TONE, &seqs);
	glbl->seqMgr().SequenceEventUnsubscribe(SequenceEvents::SEQUENCE_POSITION, &seqs);

	SequenceMenuDeinit();

	myiMS->Disconnect();
	std::cout << "Press a key to exit..." << std::endl;
	while (!_kbhit());

    return 0;
}

